from typing import Tuple


class Segment:
    """holds information about a segment of a path of an entity"""

    def __init__(
        self, origin, destination, direction, length, cumulated_length, cumulated_time
    ) -> None:
        self.origin: Tuple[float, float] = origin
        self.destination: Tuple[float, float] = destination
        self.direction: Tuple[float, float] = direction
        self.length: float = length
        self.cumulated_length: float = cumulated_length
        self.cumlated_time: float = cumulated_time



"""seize process release"""
from casymda.blocks import Delay, NamedResource, ResourceRelease, ResourceSeizeQueue
from casymda.blocks.block_components import VisualizableBlock


class SeizeProcessRelease(VisualizableBlock):
    """entities in this block follow the process:
    request resource, be processed, release resource"""

    def __init__(
        self, env, name, xy=None, ways=None, process_time=1.0, resource_capacity=1
    ):
        super().__init__(env, name, xy=xy, ways=ways)

        self.resource = NamedResource(self.env, "resource", capacity=resource_capacity)

        self.receiver = Receiver(self.env)

        self.seizer = ResourceSeizeQueue(
            env, self.name + "_seizer", resource=self.resource
        )
        self.seizer.successors.append(self.receiver)
        self.processor = Delay(env, self.name + "_processor", process_time=process_time)
        self.processor.successors.append(self.receiver)
        self.releaser = ResourceRelease(
            env, self.name + "_releaser", resource=self.resource
        )
        self.releaser.successors.append(self.receiver)

    def actual_processing(self, entity):
        yield self.env.process(self.seizer.process_entity(entity))
        yield self.env.process(self.processor.process_entity(entity))
        yield self.env.process(self.releaser.process_entity(entity))


class Receiver:
    """receiver class to act as successor for inner blocks"""

    def __init__(self, env):
        self.env = env
        self.block_resource = NamedResource(env, "receiver_resource", capacity=1)

    def process_entity(self, entity):
        """just release the resource"""
        yield self.block_resource.release(entity.block_resource_request)



"""resource release"""
from simpy import Resource

from casymda.blocks.block_components import VisualizableBlock
from casymda.blocks.entity import Entity


class ResourceRelease(VisualizableBlock):
    """releases the specified resource
    (needs to be present in the entities resources list)"""

    def __init__(
        self,
        env,
        name,
        xy=None,
        ways=None,
        resource: Resource = None,
    ):
        super().__init__(env, name, xy=xy, ways=ways)

        self.resource_to_release: Resource = resource

    def actual_processing(self, entity: Entity):

        # find request of defined resource in entity requests list
        idx = list(map(lambda x: x.resource, entity.seized_resources)).index(
            self.resource_to_release
        )
        req = entity.seized_resources[idx]

        # done immediately, but necessary for generator behavior
        yield self.resource_to_release.release(req)



"""wait for interrupt"""
from simpy import Interrupt

from casymda.blocks.block_components import VisualizableBlock


class WaitForInterrupt(VisualizableBlock):
    """processes entities until their process is interrupted"""

    def __init__(self, env, name, xy=None, ways=None):
        super().__init__(env, name, xy=xy, ways=ways)

    # note: entity.current_process might be used for interrupt
    def actual_processing(self, entity):
        try:
            yield self.env.timeout(float("inf"))
        except Interrupt:
            pass



import os
import random
from typing import Any, Dict

from . import tilemap_nx as tmnx

MATRIX_CACHE = "_temp_matrix.pickle"
LOOKUP_CACHE = "_temp_lookup.pickle"


def delete_cached(csv_path):
    paths = get_cache_paths(csv_path)
    for path in paths:
        if os.path.exists(path):
            os.remove(path)
            print("removed: " + path)
        else:
            print(path + " does not exist")


def get_cache_paths(csv_path: str):
    cache_directory = os.path.dirname(csv_path)
    pickled_matrix_path = os.path.join(cache_directory, MATRIX_CACHE)
    pickled_lookup_path = os.path.join(cache_directory, LOOKUP_CACHE)
    return pickled_matrix_path, pickled_lookup_path


class CoordinatesHolder:
    """manages info about shortest paths between certain locations from a given csv tilemap"""

    def __init__(self, csv_path: str, scale=10, include_diagonal=False) -> None:

        self.csv_path = csv_path
        self.cache_directory = os.path.dirname(csv_path)

        self.pickled_matrix_path, self.pickled_lookup_path = get_cache_paths(csv_path)

        self.scale = scale  # px / tile

        # let info get loaded by tmnx
        self.matrix_dict: Dict[Any, Any]
        self.node_lookup: Dict[Any, Any]
        self.matrix_dict, self.node_lookup = tmnx.initialize(
            self.pickled_matrix_path,
            self.pickled_lookup_path,
            self.csv_path,
            include_diagonal=include_diagonal,
        )

    def get_coords(self, target_name):
        return self.tile_to_coords(self.node_lookup[target_name])

    def get_path_coords_and_length_from_to(self, source_name, target_name):
        if source_name == target_name:
            return [self.get_coords(source_name), self.get_coords(target_name)], 0
        source, target = self.node_lookup[source_name], self.node_lookup[target_name]
        path_nodes = self.matrix_dict[source][target]["path"]
        coords_path = [self.tile_to_coords(tile) for tile in path_nodes]
        length = self.matrix_dict[source][target]["length"] * self.scale
        return coords_path, length

    def get_random_location_name(self):
        i = random.randint(0, len(self.node_lookup) - 1)
        return self.get_location_name_by_index(i)

    def get_location_name_by_index(self, index):
        return list(self.node_lookup)[index]

    def get_location_index_by_name(self, location_name):
        return list(self.node_lookup).index(location_name)

    def tile_to_coords(self, tile_tuple):
        scaled = [(c * self.scale + self.scale / 2) for c in tile_tuple]
        scaled.reverse()  # row, col => y, x
        return tuple(scaled)



from abc import abstractmethod

from casymda.blocks.block_components.state import States
from casymda.blocks.entity import Entity


class BlockVisualizer:
    """visualizer called from the visualizable_block"""

    @abstractmethod
    def animate_block(self, block, queuing_direction_x=None):
        raise NotImplementedError()

    @abstractmethod
    def change_block_state(self, block, state: States, new_value: bool):
        raise NotImplementedError()

    @abstractmethod
    def animate_entity_flow(self, entity: Entity, from_block, to_block):
        """animate_entity_flow
        destroy entity if to_block is None"""
        raise NotImplementedError()



""" process """
from casymda.blocks.block_components import VisualizableBlock


# processed entity waits for process time (timeout) before being forwarded
# to the successor


class Delay(VisualizableBlock):
    """simple delay for a given time"""

    def __init__(
        self,
        env,
        name,
        xy=None,
        ways=None,
        process_time=1.0,
        block_capacity=float("inf"),
    ):
        super().__init__(env, name, xy=xy, ways=ways, block_capacity=block_capacity)
        self.process_time = process_time

    def actual_processing(self, entity):
        yield self.env.timeout(self.process_time)



"""
2D visualization of positions on a ScaledCanvas
"""

from casymda.visualization.entity_visualizer import EntityVisualizer
from casymda.visualization.tilemap.tilemap_visualizer_interface import (
    TilemapVisualizerInterface,
)


class TilemapVisualizer(EntityVisualizer, TilemapVisualizerInterface):
    """visualizes entity positions at a given canvas"""



"""calculates and caches information on shortest paths between locations on a tilemap"""
import math
import multiprocessing
import os
import pickle
import time
from typing import Any, Dict, Tuple

import networkx as nx
from networkx.classes.function import path_weight
from numpy import genfromtxt, vectorize

# left, right, up, down / w, e, n, s (distance 1)
non_diagonal_directions = [(0, -1), (0, 1), (-1, 0), (1, 0)]
# nw, ne, se, sw (distance 1.41)
diagonal_directions = [(-1, -1), (-1, 1), (1, 1), (1, -1)]


def initialize(
    pickled_matrix_path: str,
    pickled_lookup_path: str,
    csv_path: str,
    include_diagonal=False,
) -> Tuple[Dict[Any, Any], Dict[Any, Any]]:
    """load pickled cache if present; create otherwise"""
    matrix_dict, node_lookup = try_to_load_matrix_dict_and_lookup(
        pickled_matrix_path, pickled_lookup_path
    )
    if None in (matrix_dict, node_lookup):
        matrix_dict, node_lookup = prepare_matrix_lookup(
            csv_path,
            pickled_matrix_path,
            pickled_lookup_path,
            include_diagonal=include_diagonal,
        )
    return matrix_dict, node_lookup


def prepare_matrix_lookup(
    csv_path, pickled_matrix_path, pickled_lookup_path, include_diagonal=False
):
    G = get_graph_from_csv(csv_path, include_diagonal=include_diagonal)
    matrix_dict, node_lookup = get_matrix_dict_from_csv(csv_path, G)
    sorted_node_lookup_keys = sorted(list(node_lookup))
    node_lookup = {key: node_lookup[key] for key in sorted_node_lookup_keys}
    pickle_matrix_dict_and_lookup(
        matrix_dict, pickled_matrix_path, node_lookup, pickled_lookup_path
    )
    return matrix_dict, node_lookup


def get_graph_from_csv(csv_path, valid_node_value=0, include_diagonal=False):
    array = _get_0_1_array_from_csv(csv_path)
    G = get_graph_from_tilemap_array(array, valid_node_value, include_diagonal)
    return G


def _get_0_1_array_from_csv(csv_path):
    # all values different from "1" will be marked as passable
    str_array = genfromtxt(csv_path, dtype=str, delimiter=",")
    f = vectorize(lambda x: 1 if x == "1" else 0)
    array_0_1 = f(str_array)
    return array_0_1


def get_graph_from_tilemap_array(array, valid_node_value, include_diagonal=False):

    edgelist = _get_edgelist_from_array(array, valid_node_value, include_diagonal)

    return nx.from_edgelist(edgelist)


def _get_edgelist_from_array(array, valid_node_value, include_diagonal):
    edgelist = _get_edgelist_for_directions(
        array, valid_node_value, non_diagonal_directions, 1
    )
    if include_diagonal:
        edgelist += _get_edgelist_for_directions(
            array, valid_node_value, diagonal_directions, 1.41
        )
    return edgelist


def _get_edgelist_for_directions(array, valid_node_value, directions, weight):
    edgelist = []
    for element in _get_elements_in_array(array):
        if _valid(element, array, valid_node_value):
            edges = _get_edges_for_neighbours(
                element, directions, array, valid_node_value, weight
            )
            edgelist += edges
    return edgelist


def _get_edges_for_neighbours(element, directions, array, valid_node_value, weight):
    edges = []
    row, col = element
    for direction in directions:
        neighbour = tuple(map(sum, zip(element, direction)))
        if _valid(neighbour, array, valid_node_value):
            edges.append((element, neighbour, {"weight": weight}))
    return edges


def _valid(element, array, valid_node_value):
    num_rows, num_cols = array.shape
    row, col = element
    if 0 <= row < num_rows and 0 <= col < num_cols:
        return array[element] == valid_node_value
    else:
        return False


def _get_elements_in_array(array):
    num_rows, num_cols = array.shape
    elements = []
    for row in range(num_rows):
        for col in range(num_cols):
            element = (row, col)
            elements.append(element)
    return elements


# additional method to get complete distance/path dict for all relevant nodes:


def get_matrix_dict_from_csv(csv_path_nodes, graph):
    nodes = get_relevant_nodes_from_csv(csv_path_nodes)
    nodes_list = list(nodes.keys())
    node_lookup_by_name = {value["name"]: key for key, value in nodes.items()}

    print("starting to calculate tilemap paths - this may take a while..")
    start_time = time.time()
    matrix_dict = calculate_matrix_dict(graph, nodes_list)
    print(
        "time needed for shortest paths calculation [s]: "
        + str(time.time() - start_time)
    )
    return matrix_dict, node_lookup_by_name


def get_relevant_nodes_from_csv(csv_path):
    array = genfromtxt(csv_path, dtype=str, delimiter=",")

    relevant_nodes = {}
    for element in _get_elements_in_array(array):
        if len(array[element]) != 0 and array[element] not in ["0", "1"]:
            relevant_nodes[element] = {"name": array[element]}
    return relevant_nodes


def euclidean_distance(a, b):
    return math.sqrt((b[0] - a[0]) ** 2 + (b[1] - a[1]) ** 2)


def calculate_matrix_dict(graph, nodes):
    # {from: {to1: {path:[...], length: X}, to2: ...}}
    pairs = {}  # all simple paths to calculate
    for source_node in nodes:
        for target_node in filter(lambda n: n is not source_node, nodes):
            if target_node in pairs and source_node in pairs[target_node]:
                continue
            if source_node not in pairs:
                pairs[source_node] = {}
            pairs[source_node][target_node] = None

    pairs_list_with_graph = []  # prepare data to be processed concurrently
    for source_node in pairs:
        for target_node in pairs[source_node]:
            pairs_list_with_graph.append((source_node, target_node, graph))

    mp_pool = multiprocessing.Pool(os.cpu_count())  # default
    print("multiprocessing pool size: " + str(os.cpu_count()))
    results_list = mp_pool.map(
        calculate_path_and_length_both_directions_from_tuple,
        pairs_list_with_graph,
    )

    result = {  # initialize dict
        source_node: {
            target_node: None for target_node in nodes if source_node is not target_node
        }
        for source_node in nodes
    }
    for entry in results_list:  # fill with results
        for source_node in entry:
            for target_node in entry[source_node]:
                result[source_node][target_node] = entry[source_node][target_node]
    return result


def calculate_path_and_length_both_directions_from_tuple(t):
    return calculate_path_and_length_both_directions(*t)  # node_1, node_2, graph


def calculate_path_and_length_both_directions(node_1, node_2, graph):
    path = nx.astar_path(
        graph,
        node_1,
        node_2,
        heuristic=euclidean_distance,
        weight="weight",
    )
    length = path_weight(graph, path, "weight")
    result = {
        node_1: {node_2: {"path": path, "length": length}},
        node_2: {node_1: {"path": list(reversed(path)), "length": length}},
    }
    return result


def calculate_matrix_dict_alt(graph, nodes):
    # {from: {to1: {path:[...], length: X}, to2: ...}}
    result = {}
    for source_node in nodes:
        for target_node in [n for n in nodes if n is not source_node]:
            path = nx.astar_path(graph, source_node, target_node)
            length = nx.astar_path_length(graph, source_node, target_node)
            if source_node not in result:
                result[source_node] = {}
            result[source_node][target_node] = {"path": path, "length": length}
    return result


def pickle_matrix_dict_and_lookup(matrix_dict, matrix_path, node_lookup, lookup_path):
    # tuples cannot easily be json-serialized, so we use pickle here
    for path, obj in {matrix_path: matrix_dict, lookup_path: node_lookup}.items():
        with open(path, "wb+") as f:
            pickle.dump(obj, f)


def try_to_load_matrix_dict_and_lookup(matrix_path, lookup_path):
    def load_obj(path):
        try:
            with open(path, "rb") as f:
                return pickle.load(f)
        except FileNotFoundError:
            return None

    return (load_obj(path) for path in (matrix_path, lookup_path))



import logging
import os

from casymda.visualization.web_server.sim_controller import (
    RunnableSimulation,
    SimController,
)
from flask import Flask, request, send_file, send_from_directory, Response
from flask_cors import CORS

logging.getLogger("werkzeug").setLevel(logging.ERROR)


class FlaskSimServer:
    """provides control endpoints and serves visual state of a simulation model"""

    def __init__(
        self, sim_controller: SimController, port=5000, host="0.0.0.0"
    ) -> None:
        self.sim_controller: SimController = sim_controller
        self.root_file = sim_controller.simulation.root_file
        self.port = port
        self.host = host

    def run_sim_server(self):
        app = Flask(__name__)
        CORS(app)
        app_dir = os.path.dirname(os.path.abspath(self.root_file))
        flask_dir = os.path.dirname(os.path.abspath(__file__))
        print(
            "starting flask server, port: %s, app_dir: %s, flask_dir: %s"
            % (self.port, app_dir, flask_dir)
        )

        @app.route("/")
        def root():
            # assumed to be next to this file
            HTML_FILE = "canvas_animation.html"
            return send_from_directory(flask_dir, HTML_FILE)

        @app.route("/lib-files/<filename>")
        def lib_files(filename):
            # assumed to be next to this file
            return send_from_directory(flask_dir, filename)

        @app.route("/files")
        def provide_file():
            filepath = request.args.get("filepath")
            if filepath.startswith("/") or filepath.startswith("C:"):
                # absolute path
                return send_file(filepath)
            else:
                # relative path
                return send_from_directory(app_dir, filepath)

        @app.route("/width")
        def get_width():
            return str(self.sim_controller.get_sim_width())

        @app.route("/height")
        def get_height():
            return str(self.sim_controller.get_sim_height())

        @app.route("/state")
        def get_state():
            return self.sim_controller.get_state_dumps()

        @app.route("/partial_state")
        def get_partial_state():
            return self.sim_controller.get_partial_state_dumps()

        @app.route("/state_fb")  # flatbuffers binary
        def get_state_fb():
            binary = self.sim_controller.get_state_dumps_fb()
            return Response(binary, mimetype="binary/octet-stream")

        @app.route("/partial_state_fb")  # flatbuffers binary
        def get_partial_state_fb():
            binary = self.sim_controller.get_partial_state_dumps_fb()
            return Response(binary, mimetype="binary/octet-stream")

        @app.route("/start", methods=["POST"])
        def start():
            return self.sim_controller.start_simulation_process()

        @app.route("/stop", methods=["POST"])
        def stop():
            return self.sim_controller.reset_sim()

        @app.route("/pause", methods=["POST"])
        def pause():
            return self.sim_controller.pause_sim()

        @app.route("/resume", methods=["POST"])
        def resume():
            return self.sim_controller.resume_sim()

        @app.route("/rt_factor", methods=["POST"])
        def post_rt_factor():
            value = float(request.args.get("value"))
            return self.sim_controller.set_rt_factor(value)

        # set debug=False for IDE debugging
        app.run(debug=True, threaded=False, port=self.port, host=self.host)


def run_server(runnable_sim: RunnableSimulation, port=5000, host="0.0.0.0"):
    sim_controller = SimController(runnable_sim)
    flask_sim_server = FlaskSimServer(sim_controller, port=port, host=host)
    flask_sim_server.run_sim_server()



"""source"""
from typing import Type

from casymda.blocks.block_components import VisualizableBlock
from casymda.blocks.entity import Entity


class Source(VisualizableBlock):
    """the source has a scheduled "creation_loop" process,
    which creates entities and then processes them regularly
    (as if the entity just entered a regular block,
    but with 0 time delay)"""

    def __init__(
        self,
        env,
        name,
        xy=None,
        ways=None,
        entity_type: Type[Entity] = Entity,
        inter_arrival_time: int = 0,
        max_entities: int = 5,
    ):

        super().__init__(env, name, xy=xy, ways=ways)

        self.entity_type = entity_type
        self.inter_arrival_time = inter_arrival_time
        self.max_entities = max_entities

        self.entity_counter = 0

        self.env.process(self.creation_loop())

    def creation_loop(self):
        """create entities as needed"""
        for i in range(self.max_entities):
            self.entity_counter += 1
            entity = self.entity_type(self.env, "entity_" + str(self.entity_counter))

            entity.block_resource_request = self.block_resource.request()
            yield entity.block_resource_request

            self.env.process(self.process_entity(entity))

            if i < self.max_entities - 1:
                yield self.env.timeout(self.inter_arrival_time)

    def actual_processing(self, entity):
        yield self.env.timeout(0)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuffers

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Image(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsImage(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Image()
        x.Init(buf, n + offset)
        return x

    # Image
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Image
    def Position(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from casymda.visualization.web_server.flatbuffers.CanvasPosition import (
                CanvasPosition,
            )

            obj = CanvasPosition()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Image
    def Anchor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from casymda.visualization.web_server.flatbuffers.Anchor import Anchor

            obj = Anchor()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Image
    def PhotoId(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 0

    # Image
    def Path(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Image
    def Factor(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Float32Flags, o + self._tab.Pos
            )
        return 1.0


def ImageStart(builder):
    builder.StartObject(5)


def ImageAddPosition(builder, position):
    builder.PrependStructSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(position), 0
    )


def ImageAddAnchor(builder, anchor):
    builder.PrependStructSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(anchor), 0
    )


def ImageAddPhotoId(builder, photoId):
    builder.PrependUint16Slot(2, photoId, 0)


def ImageAddPath(builder, path):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(path), 0
    )


def ImageAddFactor(builder, factor):
    builder.PrependFloat32Slot(4, factor, 1.0)


def ImageEnd(builder):
    return builder.EndObject()



"""the location of this module corresponds to
the location of the icon image files"""
import os

from casymda.blocks.block_components.state import States

CURRENT_FILE_PATH = os.path.dirname(os.path.realpath(__file__))
STATE_PATHS = {}
for state in States:
    STATE_PATHS[state] = CURRENT_FILE_PATH + "/" + state.value + ".png"


def get_file_path(for_state: States) -> str:
    """returns the file path of the state icon"""
    return STATE_PATHS[for_state]



""" entity """
from typing import List, Optional

from simpy import Environment, Resource, Process


class Entity:
    """flow object which is moved between components"""

    def __init__(self, env: Environment, name: str):
        self.env = env
        self.name = name
        self.current_process: Optional[Process] = None

        # requested resources, info needed for release
        self.seized_resources: List[Resource] = []
        self.time_of_last_arrival: float = -1

        self.block_resource_request = None

        # optional special icon path, checked by process_visualizer
        self.process_animation_icon_path: Optional[str] = None



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuffers

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Canvas(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsCanvas(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Canvas()
        x.Init(buf, n + offset)
        return x

    # Canvas
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Canvas
    def Content(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from casymda.visualization.web_server.flatbuffers.Element import Element

            obj = Element()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Canvas
    def ContentLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Canvas
    def ContentIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0


def CanvasStart(builder):
    builder.StartObject(1)


def CanvasAddContent(builder, content):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(content), 0
    )


def CanvasStartContentVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def CanvasEnd(builder):
    return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuffers

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Anchor(object):
    __slots__ = ["_tab"]

    # Anchor
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Anchor
    def X(self):
        return self._tab.Get(
            flatbuffers.number_types.Float32Flags,
            self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(0),
        )

    # Anchor
    def Y(self):
        return self._tab.Get(
            flatbuffers.number_types.Float32Flags,
            self._tab.Pos + flatbuffers.number_types.UOffsetTFlags.py_type(4),
        )


def CreateAnchor(builder, x, y):
    builder.Prep(4, 8)
    builder.PrependFloat32(y)
    builder.PrependFloat32(x)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: flatbuffers

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Element(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAsElement(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Element()
        x.Init(buf, n + offset)
        return x

    # Element
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Element
    def Id(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint16Flags, o + self._tab.Pos
            )
        return 0

    # Element
    def ElementType(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Element
    def Element(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            from flatbuffers.table import Table

            obj = Table(bytearray(), 0)
            self._tab.Union(obj, o)
            return obj
        return None


def ElementStart(builder):
    builder.StartObject(3)


def ElementAddId(builder, id):
    builder.PrependUint16Slot(0, id, 0)


def ElementAddElementType(builder, elementType):
    builder.PrependUint8Slot(1, elementType, 0)


def ElementAddElement(builder, element):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(element), 0
    )


def ElementEnd(builder):
    return builder.EndObject()



"""canvas"""

from abc import abstractmethod


class ScaledCanvas:
    """overview of canvas methods called from visualizer"""

    def __init__(self, scale: float) -> None:
        self.scale = scale

    @abstractmethod
    def load_image_file(self, path: str):
        """takes file path and returns loaded image file reference"""

    @abstractmethod
    def create_image(self, x_coord: int, y_coord: int, image_file, anchor="c") -> int:
        """places image on canvas and returns image reference"""

    @abstractmethod
    def create_text(
        self,
        x_coord: int,
        y_coord: int,
        text: str = "",
        anchor: str = "se",
        fill: str = "black",
        font: str = "Helvetica 16",
    ) -> int:
        """places text on canvas and returns reference"""

    @abstractmethod
    def delete(self, element_id: int) -> bool:
        """deletes element and returns True if deleted"""

    @abstractmethod
    def set_coords(self, element_id: int, x_y: tuple):
        """places the given element at a given position"""

    @abstractmethod
    def set_text_value(self, element_id: int, text: str):
        """changes the elements text"""

    @abstractmethod
    def get_width(self) -> int:
        """returns width"""

    @abstractmethod
    def get_height(self) -> int:
        """returns heigth"""

    def _scale_coords(self, x_y: tuple):
        """returns tuple values multiplied by factor"""
        return tuple(map(lambda c: int(c * self.scale), x_y))



import json
import pathlib
from operator import sub
from typing import Dict, List

import black
import xmltodict

from casymda import __version__

print("casymda version: " + __version__)

OFFSET_REDUCTION = (5, 3)

MAPPING = {
    "dataStoreReference": "resources",
    "startEvent": "sources",
    "endEvent": "sinks",
    "task": "tasks",
    "exclusiveGateway": "xGateways",
    "sequenceFlow": "connectors",
    "textAnnotation": "textAnnotations",
    "association": "associations",
}

COMPONENTS = ["sources", "sinks", "tasks", "xGateways"]


def parse_bpmn(
    bpmn_path: str, json_path: str, template_path: str, model_path: str
) -> None:
    """
    this method reads from a given (camunda-modeler) bpmn/xml file and
    completes a given skeleton template by initialized blocks and
    their connections
    """

    bpmn = _to_json_dict(bpmn_path, json_path)

    elements = {
        value: _get_elements_of_type(bpmn, key) for key, value in MAPPING.items()
    }

    components = _get_components_from_elements(elements)

    _add_class_and_id_names(components + elements["resources"])

    insertions = _generate_insertions(bpmn, elements, components)
    model_components = _generate_model_components(components)
    model_graph_names = _generate_model_graph_names(elements, components)

    _write_model_file(
        model_path, template_path, insertions, model_components, model_graph_names
    )

    path = pathlib.Path(model_path)
    file_mode = black.FileMode()
    black.format_file_in_place(path, False, file_mode, write_back=black.WriteBack.YES)

    print("\nBPMN parsed.\n")


def _generate_model_components(components):
    return (
        "self.model_components = {"
        + ", ".join(
            [
                "'" + comp["id_name"] + "': self." + comp["id_name"]
                for comp in components
            ]
        )
        + "}"
    )


def _generate_model_graph_names(elements, components):
    return {
        comp["id_name"]: _get_successor_names_list(comp, elements, components)
        for comp in components
        if not (_get_successor_names_list(comp, elements, components) is None)
    }


def _add_class_and_id_names(elements: list):
    for element in elements:
        names = element["@name"].split(":")
        # add removal of possibly leading/trailing blanks
        names = [n.strip() for n in names]
        element["class_name"] = names[0]
        element["id_name"] = names[1]


def _get_components_from_elements(elements: Dict[str, List[str]]) -> List[str]:
    components: List[str] = []
    for component_label in COMPONENTS:
        components += elements[component_label]
    return components


def _generate_insertions(bpmn, elements, components):
    insertions = []
    offset = _get_offset(bpmn)
    for comp in elements["resources"] + components:
        class_name = comp["class_name"]
        id_name = comp["id_name"]

        x_y = _get_shape_xy(comp["@id"], bpmn, offset=offset)

        res_name = _get_related_resource_id_name(comp, elements)

        if res_name is None:
            insertions.append(
                "self.%s = %s(self.env, '%s', xy=%s"
                % (id_name, class_name, id_name, x_y)
            )
        else:
            insertions.append(
                "self.%s = %s(self.env, '%s', resource=self.%s, xy=%s"
                % (id_name, class_name, id_name, res_name, x_y)
            )

        insertions[-1] += _get_related_text_annotations(comp, elements)

        insertions[-1] += ", ways=" + str(
            _get_ways_to_successors(comp, bpmn, elements, components, offset=offset)
        )

        # close bracket and 2x newline
        insertions[-1] += ")\n\n"
    return insertions


def _write_model_file(
    model_path, template_path, insertions, model_components, model_graph_names
):
    # read template file
    with open(template_path, "r") as template_file:
        template = template_file.readlines()

    loc = 0

    # find resources+components insertion space
    for idx, line in enumerate(template):
        if "#!resources+components" in line:
            loc = idx + 1
            template.insert(loc, "\n")
            break

    for i_line in insertions:
        loc += 1
        template.insert(loc, "".join([" " for _ in range(8)]) + i_line)

    # find #!model insertion space
    for idx, line in enumerate(template):
        if "#!model" in line:
            loc = idx
            break

    loc += 1
    template.insert(loc, "\n" + "".join([" " for _ in range(8)]) + model_components)
    loc += 1
    template.insert(
        loc,
        "\n\n"
        + "".join([" " for _ in range(8)])
        + "self.model_graph_names = "
        + json.dumps(model_graph_names),
    )

    with open(model_path, "w+") as model_file:
        model_file.writelines(template)


def _to_json_dict(bpmn_path: str, json_path: str) -> dict:
    """save bpmn xml as json and return dictionary"""
    with open(bpmn_path, "rb") as bpmn_file:
        bpmn: dict = xmltodict.parse(bpmn_file)
        bpmn = json.loads(json.dumps(bpmn, indent=4))
    with open(json_path, "w+") as json_file:
        json.dump(bpmn, json_file, indent=4)
    return bpmn


def _get_offset(xmlo):
    offset = [float("inf"), float("inf")]

    coords_list = []

    for shape in xmlo["bpmn:definitions"]["bpmndi:BPMNDiagram"]["bpmndi:BPMNPlane"][
        "bpmndi:BPMNShape"
    ]:
        # find all bounded elements and include their label bounds
        shape_bounds = [shape["dc:Bounds"]]
        if "bpmndi:BPMNLabel" in shape:
            shape_bounds.append(shape["bpmndi:BPMNLabel"]["dc:Bounds"])
        for bounds in shape_bounds:
            # look for global min
            coords_list.append((float(bounds["@x"]), float(bounds["@y"])))

    for edge in xmlo["bpmn:definitions"]["bpmndi:BPMNDiagram"]["bpmndi:BPMNPlane"][
        "bpmndi:BPMNEdge"
    ]:
        for waypoint in edge["di:waypoint"]:
            coords_list.append((float(waypoint["@x"]), float(waypoint["@y"])))

    for coords in coords_list:
        for idx, val in enumerate(offset):
            if coords[idx] < val:
                offset[idx] = int(coords[idx])

    offset = tuple(map(sub, offset, OFFSET_REDUCTION))

    return offset


def _get_shape_xy(elid, xmlo, offset=(0, 0)):
    """find position of a shape by element id"""
    for shape in xmlo["bpmn:definitions"]["bpmndi:BPMNDiagram"]["bpmndi:BPMNPlane"][
        "bpmndi:BPMNShape"
    ]:
        if shape["@bpmnElement"] == elid:
            x_y = [
                float(shape["dc:Bounds"]["@x"])
                + int(float(shape["dc:Bounds"]["@width"]) / 2),
                float(shape["dc:Bounds"]["@y"])
                + int(float(shape["dc:Bounds"]["@height"]) / 2),
            ]
            x_y = tuple(map(int, map(sub, x_y, offset)))
            return x_y
    return None


def _get_elements_of_type(bpmn, element_type) -> List[str]:
    process = bpmn["bpmn:definitions"]["bpmn:process"]
    if "bpmn:" + element_type in process:
        elements = process["bpmn:" + element_type]
        return elements if isinstance(elements, list) else [elements]
    return []


def _contains_resource_relation(component):
    if ("bpmn:dataInputAssociation" in component) or (
        "bpmn:dataOutputAssociation" in component
    ):
        return True
    return False


def _find_res_id_name(bpmn_id, elements):
    for res in elements["resources"]:
        if res["@id"] == bpmn_id:
            return res["id_name"]
    return None


def _get_related_resource_id_name(component, elements):
    if "bpmn:dataInputAssociation" in component:
        bpmn_id = component["bpmn:dataInputAssociation"]["bpmn:sourceRef"]
        return _find_res_id_name(bpmn_id, elements)
    if "bpmn:dataOutputAssociation" in component:
        bpmn_id = component["bpmn:dataOutputAssociation"]["bpmn:targetRef"]
        return _find_res_id_name(bpmn_id, elements)
    return None


def _get_related_text_annotations(component, elements):
    associations = [
        assoc
        for assoc in elements["associations"]
        if assoc["@sourceRef"] == component["@id"]
    ]
    # get corresponding textAnnotations
    annotations = []
    for assoc in associations:
        annotations += [
            annot
            for annot in elements["textAnnotations"]
            if annot["@id"] == assoc["@targetRef"]
        ]
    params = []
    for annot in annotations:
        params += [x.strip() for x in annot["bpmn:text"].split(";")]
    text = ""
    if len(params) > 0:
        text = ", " + ", ".join(params)
    return text


def _get_outgoing_sequence_flows(comp):
    # get list of outgoing sequence flows
    if "bpmn:outgoing" not in comp:
        return None
    outgoing_sfs = comp["bpmn:outgoing"]
    outgoing_sfs = (
        [outgoing_sfs] if not isinstance(outgoing_sfs, list) else outgoing_sfs
    )
    return outgoing_sfs


def _get_successor_names_list(comp, elements, components):
    outgoing_sfs = _get_outgoing_sequence_flows(comp)
    res = []
    if outgoing_sfs is None:
        return res
    for seq_flow in outgoing_sfs:
        # find SF element -> find targetRef -> append id_name
        for con in elements["connectors"]:
            if con["@id"] == seq_flow:
                target_id = con["@targetRef"]
                for target_comp in components:
                    if target_comp["@id"] == target_id:
                        res.append(target_comp["id_name"])
                        break
                break
    return res


def _get_diagram_waypoints_for_sf_id(sf_id, xmlo, offset=(0, 0)):
    waypoints = []
    for edge in xmlo["bpmn:definitions"]["bpmndi:BPMNDiagram"]["bpmndi:BPMNPlane"][
        "bpmndi:BPMNEdge"
    ]:
        if edge["@bpmnElement"] == sf_id:
            for waypoint in edge["di:waypoint"]:
                waypoints.append(
                    tuple(
                        map(
                            sub,
                            (int(float(waypoint["@x"])), int(float(waypoint["@y"]))),
                            offset,
                        )
                    )
                )
            return waypoints
    return None


def _get_ways_to_successors(comp, xmlo, elements, components, offset=(0, 0)):
    results = {}  # {"suc name":[{x: int, y: int}, ...]}
    outgoing_sfs = _get_outgoing_sequence_flows(comp)
    if outgoing_sfs is not None:
        for seq_flow in outgoing_sfs:
            # find SF element -> find targetRef -> append id_name: waypoints
            for con in elements["connectors"]:
                if con["@id"] == seq_flow:
                    target_id = con["@targetRef"]
                    for target_comp in components:
                        if target_comp["@id"] == target_id:
                            waypoints = _get_diagram_waypoints_for_sf_id(
                                con["@id"], xmlo, offset=offset
                            )
                            # [{x: int, y: int}]
                            results[target_comp["id_name"]] = waypoints
                            break
                    break
    return results



"""gateway"""
from casymda.blocks.block_components import VisualizableBlock
from casymda.blocks.entity import Entity


class RoundRobinGateway(VisualizableBlock):
    """gateway which cyclically chooses next successor for incoming entities"""

    def __init__(self, env, name, xy=None, ways=None):
        super().__init__(env, name, xy=xy, ways=ways)

        self.round_robin = self.round_robin_generator()

    def actual_processing(self, entity):
        yield self.env.timeout(0)

    def find_successor(self, entity: Entity):
        return next(self.round_robin)

    def round_robin_generator(self):
        """iterate over successors, remember last one"""
        last = -1
        while True:
            last += 1
            last %= len(self.successors)
            yield self.successors[last]



"""convenience import for easier access"""
from casymda import __version__
from casymda.blocks.delay import Delay
from casymda.blocks.entity import Entity
from casymda.blocks.named_resource import NamedResource
from casymda.blocks.resource_release import ResourceRelease
from casymda.blocks.resource_seize_queue import ResourceSeizeQueue
from casymda.blocks.round_robin_gateway import RoundRobinGateway
from casymda.blocks.seize_process_release import SeizeProcessRelease
from casymda.blocks.sink import Sink
from casymda.blocks.source import Source
from casymda.blocks.wait_for_interrupt import WaitForInterrupt
from casymda.blocks.tilemap.tilemap_movement import TilemapMovement
